=begin
###
Copyright 2011-2017 TIG (c)
All rights reserved.
THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
###
Name:

  TIG-AreaTextTag.rb
  
Description:

  NOTE from v1.1 this tool loads as an extension and might need activating 
  using Window > Preferences > Extensions > AreaTextTag when it is first 
  installed.  Thereafter it will load until you deactivate it.
  
  It adds a submenu to the Plugins menu OR a Toolbar of the same name 
  which needs activating from View > Toolbars [after the Extension is 
  loaded]:-
  
'AreaTextTag Tools...' containing...
  
'Tag Picked Faces'
  This lets you add a special 'AreaTextTag' onto a Face by Picking it. 
  This 'Tag' is a component flat 3d Text, colored 'black', although you 
  can recolor it later as desired.  
  This 'Tag' is glued to the picked Face, centered on the picked point 
  Because it is glued the 'Tag' moves with the Face.
  The 'Tag' can be manually relocated anywhere on the Face's plane.
  By default the area in the 'Tag' is formatted in 'm2' or 'ft2' 
  depending on the SKP's units.
  The 'm2' value is rounded to 2 dp [12.34], and the 'ft2' value is 
  rounded to 1 dp [1.2].
  These 'units' and 'dps' can be reset using the 'Change Settings...' 
  menu option.
  By default the 'Tag' is simply named "1.23 m2" [assuming that's the 
  area/units], upon making it a  dialog asks for the Tag's 'text', press 
  Cancel to leave it as "1.23 m2", or enter additional text and press 
  OK, the text/area updates with the new values.
  NOTE: that there are several tricks to formatting the Tag's text:
  You enter/reuse/leave <> where the 'area' value is to go.
  updated area text will be added later, therefore text without a <> has 
  one appended so that the area is appended to it [with a units suffix].
  Entering: <> or leaving the text blank OR pressing Cancel gives 
    "1.23 m2"
  Using: <-> gives
    "1.23" [i.e. no units suffix is shown]
  Entering: Office 666 <> OR Office 666 [WITHOUT the <>] 
    "Office 666 1.23 m2"
  Entering: Office 666 <->
    "Office 666 1.23" [i.e. no units suffix is shown]
  You can also add text after the 'area' thus:
  <> Office 999
    "1.23 m2 Office 999"
  or
  Office 666 <> (Admin)
    "Office 666 1.23 m2 (Admin)"
  You can also separate the text over two or more lines by including 
  '\n' to denote the newlines, e.g:
  Office 666\n<>
    "Office 666
    1.23 m2"
		
	From v3.0 you can also include a 'Perimeter' in the Tag.
	Use the special character combination ||
	
	Tiling Area <> Perim ||
	
	Which would then display a tagged 1 m2 area thus:
	
		"Tiling Area 1.00 m2 Perim 4.00 m"
		
	The Perimeter is always given in the units / dp eqivalent to the area format.
	Note that areas in Acres use ft and in Hectares use m, for their perimeters.
	
	To suppress the Perimeter's units use the combination |-|
	
  Note that the limitation on the characters that can be used in 
  the Text, '<>' and '<->' which determine where the numeric 'area' comes.  
	Note that from v3.0 || and |-| are reserved for the 'perimeter' comes
  Commas ',' are best avoided if CSV files are anticipated as they are 
  replaced with a ';' anyway.
	
  The Tool lets you place Tags until you press <Esc> or choose another 
  Tool to exit.
  Later on, if desired you can Edit a 'Tag' - to change the its 'Text' - 
  e.g. to alter its associated room/area name - Select it and use the 
  appropriate 'Edit Selected Tag' tool.
  If the associated Face is changed then you should Update the Tag so 
  that the reported area will be updated automatically for you.
  If a Face is erased and it had an 'Tag' associated with it then that 
  Tag returns an error if you try to Edit it.
  Note that AreaTextTags should always be placed in the same context as 
  the Face itself - in fact the tool prevents you selecting a face 
  outside of the active_entities.
  From v1.3 there is an option to report 'Surface Areas'.  If this is 
  set then a Face forming part of a larger mesh Surface [soft edges]
  shows the area of the Surface rather than that of the Face itself.  
  Remember not to pick more than one Face within a mesh as you'll get 
  duplicate Tags.

'Tag Selected Faces'
  Similar to 'Tag Picked Faces' but you Preselect the faces to be tagged 
  and there is no option to add additional text to the Tag, although you 
  can Edit a Tag later to add this.  Also the Tags are glued on the 
  center of each Face, although these can be Moved later if desired.
  Note that if there is no Selection you are give the option to Tag 
  ALL Faces in the active_entities...
  From v1.3 there is an option to report 'Surface' areas.  If this is 
  set then a Face forming part of a larger mesh Surface [soft edges]
  shows the area of the Surface rather than that of the Face itself. 
  Because Selected Faces might belong to the same Surface, then only one 
  Face per Surface will be filtered from the Selected Faces.
  To maintain control over which Face in a Surface gets the Tag added 
  the 'Tag Picked Faces' option is recommended [see above].

'Edit Selected Tag'
  This opens a dialog in which you can change the 'Text' part of a 
  Selected AreaTextTag.  It also updates the Tag to use the face's 
  current-area; however, the current Tag's units etc are retained, so to 
  update these to have the current-settings use the 'Update' tool...
  
'Change Settings + Tag Picked Faces'
  This opens a dialog in which you can change the units that the 
  'Area Text Tag' will use from then on, drop a drop-down list, thus:
  'Area Units:'
    in2
    ft2***
    yds2
    acres
    mm2
    cm2
    m2***
    hectares
	  m2(ft2)
  The ones marked with *** are the automatic defaults for 
  imperial/metric SKP models.
  The m2(ft2) option is for the UK where areas are measured in m2 BUT 
  commercial floorspace is rented in m2/ft2 !  With that option the dp is
  overridden and m2 are reported to 1 dp and ft2 to 0 dp*** thus:
   1.0 m2
  (11 ft2)
  The 'current units' etc are remembered within the saved model SKP, 
  across sessions.
  In this dialog you can also change the Area Tag's 
  'Text DP:'     - default 2 for m2 and 1 for ft2***
  'Text Font:'   - default 'Arial'
				         [some fonts might be unavailable on your OS]
	'Text Alignment:'- default 'Center'- alternatives Left|Right [from v3.0]
  'Text Height:'   - default 100mm/4"
  
  From v1.3 there's an option to
  'Report 'Surface' Areas:' - default 'No'
  If this is set to 'Yes' then any Face within a mesh Surface [soft 
  edges] is given a Tag that reports the Surface area, rather than just 
  the area of the Face itself.
  
  On OK the 'Tag Picked Faces' Tool then starts, it will now use the 
  'current-settings'...
  
'Update Selected Tags'
  All selected 'AreaTextTags' are updated
  You are then asked if you want to update:
  AREAS ONLY [=Yes]
  or
  AREAS AND ALL OTHER SETTINGS [=No]
  i.e. to use the 'current-settings' - units, dps, font, height.
  The current area of each Tag's 'glued_to face' is ALWAYS updated 
  within the Text of the Tag.
  The Selection can contain other non-Tags, which will then be ignored.
  Note that if there is no Selection you are give the option to Update 
  ALL Tags...
  From v1.3 there's an option to 'Report Surface Areas:' if it's set to 
  'Yes' then any Face within a mesh Surface [soft edges] has a Tag that 
  reports the Surface's area, rather than just the area of the Face 
  itself.  So any Face that has been changed since it was last Tagged, 
  or its Tag was updated, will also change the reported area - 
  e.g. Faces might have been added to [or removed from] a mesh, or the 
  Face itself might be not even be part of a mesh any longer.

'Export Tags' to CSV'
  You must save a new model before running this tool.
  It processes Tags in the current active context.
  All Tags have their areas auto-Updated, then this tool runs.
  Note that from v1.3 Faces that are part of a Surface will report the 
  Surface's are IF the current-settings specify this.
  Any Tags that are NOT glued to a Face are NOT Reported [the closing 
  dialog tells this and highlights them].
  Any Tags that are 'duplicated' on a Face are NOT Reported [the closing 
  dialog tells this and highlights them].
  It then makes a report of the Tags in CSV format, listing them by:
    LAYER, TEXT, PERIMETER, AREA, UNITS
  Reporting the 'Layer' helps divide the results up, so putting your 
  Tags onto Layers named 'Level0', 'Level1' etc helps you use the report 
  more easily.
  The reported Perimeter/Area is in each Tag's 'units' - typically m/m2 
  or ft/ft2 - however, areas in acres have perimeters is ft and if in 
  hectares perimeters are in meters.
  If the units are 'm2(ft2)' two extra columns are added for AREA2 and 
  UNITS2 to show the m2 AND ft2 version.
  Any commas [,] in reported Text/Layers are changed into colons [;] to 
  avoid formatting errors in the CSV.
  If the Model name is XXXX the report is written into 
    XXXX_AreaTextTagReport.csv.  
  The CSV report is made in the Model's folder.  
  If that Report is already open [>>>error] you are told/warned.
  Otherwise the CSV is written and on completion the CSV file is opened.
  A CSV file is readable by Excel and similar programs.
  It can be linked into a XLS file that has 'formatting' so then if the 
  CSV updates the pre-formatted linked XLS file will change to suit.
  
###
=end
#####################

module TIG
module AreaTextTag

### menu
unless file_loaded?(__FILE__)
  ### commands
  cmd1=UI::Command.new("Tag Picked Faces"){Sketchup.active_model.select_tool(self::TagPicked.new())}
		cmd1.tooltip='Tag Picked Faces'
    cmd1.status_bar_text='AreaTextTag: Tag Picked Faces...'
    cmd1.small_icon=File.join(File.dirname(__FILE__), 'AreaTextTag-TagPicked-16x16.png')
    cmd1.large_icon=File.join(File.dirname(__FILE__), 'AreaTextTag-TagPicked-24x24.png')
	###
  cmd2=UI::Command.new("Tag Selected Faces"){self::tagSelected()}
		cmd2.tooltip='Tag Selected Faces'
    cmd2.status_bar_text='AreaTextTag: Tag Selected Faces...'
    cmd2.small_icon=File.join(File.dirname(__FILE__), 'AreaTextTag-tagSelected-16x16.png')
    cmd2.large_icon=File.join(File.dirname(__FILE__), 'AreaTextTag-tagSelected-24x24.png')
	###
  cmd3=UI::Command.new("Edit Selected Tag"){self.edit()}
		cmd3.tooltip='Edit Selected Tag'
    cmd3.status_bar_text='AreaTextTag: Edit Selected Tag...'
    cmd3.small_icon=File.join(File.dirname(__FILE__), 'AreaTextTag-edit-16x16.png')
    cmd3.large_icon=File.join(File.dirname(__FILE__), 'AreaTextTag-edit-24x24.png')
	###
  cmd4=UI::Command.new("Change Settings + Tag Picked Faces"){self.settings()}
		cmd4.tooltip='Change Settings'
    cmd4.status_bar_text='AreaTextTag: Change Settings + Tag Picked Faces...'
    cmd4.small_icon=File.join(File.dirname(__FILE__), 'AreaTextTag-settings-16x16.png')
    cmd4.large_icon=File.join(File.dirname(__FILE__), 'AreaTextTag-settings-24x24.png')
	###
  cmd5=UI::Command.new("Update Selected Tags"){self.update()}
		cmd5.tooltip='Update Selected Tags'
    cmd5.status_bar_text='AreaTextTag: Update Selected Tags...'
    cmd5.small_icon=File.join(File.dirname(__FILE__), 'AreaTextTag-update-16x16.png')
    cmd5.large_icon=File.join(File.dirname(__FILE__), 'AreaTextTag-update-24x24.png')
	###
  cmd6=UI::Command.new("Export Tags to CSV"){self.report()}
		cmd6.tooltip='Export Tags to CSV'
    cmd6.status_bar_text='AreaTextTag: Export Tags to CSV...'
    cmd6.small_icon=File.join(File.dirname(__FILE__), 'AreaTextTag-report-16x16.png')
    cmd6.large_icon=File.join(File.dirname(__FILE__), 'AreaTextTag-report-24x24.png')
	###
  menu=UI.menu("Plugins").add_submenu("AreaTextTag Tools...")
		menu.add_item(cmd1)
		menu.add_item(cmd2)
    menu.add_item(cmd3)
    menu.add_item(cmd4)
    menu.add_item(cmd5)
    menu.add_item(cmd6)
	###
	toolbar=UI::Toolbar.new('AreaTextTag Tools...')
    toolbar.add_item(cmd1)
    toolbar.add_item(cmd2)
    toolbar.add_item(cmd3)
    toolbar.add_item(cmd4)
    toolbar.add_item(cmd5)
    toolbar.add_item(cmd6)
    toolbar.restore if toolbar.get_last_state.abs==1 #TB_VISIBLE/NEVER
end
file_loaded(__FILE__)
###
### add observers for auto-update of areas... FUTURE...
### check only ever one instance of an AreaTextTag#

### check on def change that instance reports area of face it's glued to

#################
class self::TagPicked

def initialize()
  @model=Sketchup.active_model
  @ents=@model.active_entities
  ### get settings units/dp/font/height***
  @units=@model.get_attribute("AreaTextTag","units","m2")
  @model.set_attribute("AreaTextTag","units",@units)
  @dp=@model.get_attribute("AreaTextTag","dp","2")
  @model.set_attribute("AreaTextTag","dp",@dp)
  @font=@model.get_attribute("AreaTextTag","font","Arial")
  @model.set_attribute("AreaTextTag","font",@font)
	### v3.0***
	@align=@model.get_attribute("AreaTextTag","align","Center")
  @model.set_attribute("AreaTextTag","align",@align)
	###
  if @model.options["UnitsOptions"]["LengthUnit"]<=1 ### it's '/"
    high=4.inch
  else
    high=100.mm
  end
  @height=@model.get_attribute("AreaTextTag","height",high)
  @model.set_attribute("AreaTextTag","height",@height)
  @surf=@model.get_attribute("AreaTextTag","surface","No")
  @model.set_attribute("AreaTextTag","surface",@surf)
  ###
  @surfaces=[]
  ###
end#initialize

def activate
  ###
  @ip=Sketchup::InputPoint.new
  @msg="AreaTextTag: New Tag - Pick Face..."
  Sketchup::set_status_text(@msg)
  @face=nil
  ###
end#activate

def deactivate(view)
  view.invalidate
  return nil
end#deactivate

def onCancel(reason, view)
  view.invalidate
  Sketchup.send_action("selectSelectionTool:")
  return nil
end#oncancel
    
def resume(view)
  Sketchup::set_status_text(@msg)
end#resume

def onLButtonUp(flags,x,y,view)
  @ip.pick(view,x,y)
  if @ip.valid?
    @pt=@ip.position
  end#if
  @ip.clear
  ph=view.pick_helper
  ph.do_pick(x,y)
  bestf=ph.picked_face
  if bestf && bestf.parent.entities==@ents
    @face=bestf
		if @surfaces.include?(@face)
			UI.messagebox("Area Text Tag:\n\nFace in already Tagged !")
			@face=nil
		else
			self.add_text()
		end
  else
    UI.messagebox("Area Text Tag:\n\nPick a Face in the active_entities !")
		@face=nil
  end#if
end#onLButtonUp

##########################
def surface_data(face=nil)
  return [0, 0, [nil]] unless face && face.valid? && face.is_a?(Sketchup::Face)
  unless face.edges.any?{|e|e.soft?} ### NOT a surface
		peri=0
		face.edges.each{|e|peri+=e.length} ### return perimter anyway...
		return [face.area, peri, [face]]
  end
  area=0
  peri=0
  faces=self.get_surface(face)
  faces.each{|f|
    area+=f.area
		pe=0
		f.edges.each{|e|
			next if e.soft?
			next if e.faces[1] && faces.include?(e.faces[0]) && faces.include?(e.faces[1])
			pe+=e.length
		}
		peri+=pe
  }
  return [area, peri, faces]
end#surface_data
###
def get_surface(face)### thanks to thomthom
  surface={} ### start with hash for speed
  stack=[face]
  until stack.empty?
    stack.shift.edges.select{|e|e.soft?}.each{|edge|
      edge.faces.each{|fac|
        next if surface.key?(fac)
        stack << fac
        surface[fac]=fac
      }
    }
  end
  return surface.keys ### array
end#get_surface
######################

def add_text()
  return nil unless @face
  if @surfaces.include?(@face)
    UI.beep
    return nil
  end
  ###
  @model.start_operation("Area Text Tag")
  ###
  @surfaces << @face
  area, perim, faces = self.surface_data(@face)
	@surfaces << faces
	@surfaces.flatten!
	@surfaces.compact!
  @surfaces.uniq!
  ###
  area2=nil
	area2ever=false
  case @units
    when "ft2"
			area=area/144
			perim=perim/12
			punit="ft"
    when "yds2"
			area=area/1296
			perim=perim/36
			punit="yds"
    when "acres"
			area=area*0.000000159422507907
			perim=perim/12
			punit="ft"
    when "mm2"
			area=area*645.16
			perim=perim*25.4
			punit="mm"
    when "cm2"
			area=area*6.4516
			perim=perim*2.54
			punit="cm"
    when "m2"
			area=area*0.00064516
			perim=perim*0.0254
			punit="m"
    when "hectares"
			area=area*0.000000064516
			perim=perim*0.0254
			punit="m"
		when "in2"
			area=area
			perim=perim
			punit="in"
		when "m2(ft2)"
			area=area*0.00064516
			perim=perim*0.0254
			punit="m"
			area2=area*10.7639104167097223
			area2ever=true
		else
			area=0
			perim=0
			punit=""
   end
  atxt=sprintf("%."+@dp+"f", area)
  atxtu=atxt+" "+(@units.gsub(/2/,"²"))
  if area2 ### "m2(ft2)"
    atxt=sprintf("%.1f", area)
    atxtu=atxt+" m²"
		atxt2=sprintf("%.0f", area2)
		atxtu2=atxt2+" ft²"
		atxt=atxt+"\n("+atxt2+")"
		atxtu=atxtu+"\n("+atxtu2+")"
  end
  #ttt=atxtu
  #begin
    #atxtu=atxtu.unpack("U*").map{|c|c.chr}.join
  #rescue
    #atxtu=ttt
  #end
	case @align
		when "Left"
			align = TextAlignLeft
		when "Right"
			align = TextAlignRight
		else ### Center
			align = TextAlignCenter
	end
  ###
  gp=@ents.add_group()
  gents=gp.entities
  gents.add_3d_text(atxtu, align, @font, false, false, @height, 0, 0, true, 0)
	gents.to_a.each{|e|
    if e.is_a?(Sketchup::Face)
			e.reverse! if e.normal.z<0
		elsif e.is_a?(Sketchup::Edge)
			e.hidden=true if @hidden=="Yes"
		end
		e.casts_shadows=false
    e.receives_shadows=false
  }
  ###
  ct=gp.bounds.center
  tr=Geom::Transformation.translation(ct.vector_to([0,0,1.mm]))
  gents.transform_entities(tr, gents.to_a)
  ###
  ins=gp.to_component
  ins.material="Black"
  defn=ins.definition
  dents=defn.entities
  defn.name="AreaTextTag#1"
  defn.behavior.is2d=true
  defn.behavior.snapto=SnapTo_Arbitrary
  ###
  ins.casts_shadows=false
  ins.receives_shadows=false
  ###
  ### relocate to pt
  tr=Geom::Transformation.new(@pt)
  ins.transform!(tr)
  ### rotate onto face
  norm=@face.normal
  if norm==Z_AXIS.reverse
    tr=Geom::Transformation.rotation(@pt, X_AXIS, 180.degrees)
    ins.transform!(tr)
  elsif norm!=Z_AXIS
    ### flatten norm
    p1=ORIGIN.offset(norm)
		p1.z=0
		fvec=ORIGIN.vector_to(p1)
			ang=fvec.angle_between(ins.transformation.yaxis.reverse)
		ang= -ang if fvec.x<0
		tr=Geom::Transformation.rotation(@pt, Z_AXIS, ang)
			ins.transform!(tr)
		### flip onto face
		ang=norm.angle_between(Z_AXIS)
		tr=Geom::Transformation.rotation(@pt, ins.transformation.xaxis, ang)
    ins.transform!(tr)
  end
  ### set gluing to face etc 
  ins.glued_to=@face
  ###
  ### finalize text content
  text=self.dialog()
	###
  if text==nil || text==""
		text=""
    txt=atxtu
  elsif text=~/[<][>]/
    txt=text.gsub(/[<][>]/, atxtu)
  elsif text=~/[<][-][>]/
    txt=text.gsub(/[<][-][>]/, atxt)
  else
    txt=text
  end
	### perimeter
	if txt=~/[|][|]/
		ptxt=sprintf("%."+@dp+"f", perim)
		ptxt<<" "+punit
		txt.gsub!(/[|][|]/, ptxt)
	elsif txt=~/[|][-][|]/
		ptxt=sprintf("%."+@dp+"f", perim)
		txt.gsub!(/[|][-][|]/, ptxt)
	end
  ###
  ### trap newlines \n
  txt.gsub!(/[\\]n/,"\n")
  #ttt=txt
  #begin
    #txt=txt.unpack("U*").map{|c|c.chr}.join
  #rescue
    #txt=ttt
  #end
	case @align
		when "Left"
			align = TextAlignLeft
		when "Right"
			align = TextAlignRight
		else ### Center
			align = TextAlignCenter
	end
  dents.erase_entities(dents.to_a)
  dents.add_3d_text(txt, align, @font, false, false, @height, 0, 0, true, 0)
  ###
  bb=Geom::BoundingBox.new
  dents.each{|e|bb.add(e.bounds)}
  ct=bb.center
  tr=Geom::Transformation.translation(ct.vector_to([0,0,1.mm]))
  dents.transform_entities(tr, dents.to_a)
  ###
  dents.to_a.each{|e|
    if e.is_a?(Sketchup::Face)
			e.reverse! if e.normal.z<0
		elsif e.is_a?(Sketchup::Edge)
			e.hidden=true if @hidden=="Yes"
		end
		e.casts_shadows=false
    e.receives_shadows=false
  }
  defn.invalidate_bounds
  ### remember dialog's text etc as attributes
  defn.set_attribute("AreaTextTag","text",text)
  defn.set_attribute("AreaTextTag","units",@units)
  defn.set_attribute("AreaTextTag","dp",@dp)
  defn.set_attribute("AreaTextTag","font",@font)
  defn.set_attribute("AreaTextTag","align",@align)
  defn.set_attribute("AreaTextTag","height",@height)
  defn.set_attribute("AreaTextTag","hidden",@hidden)
  ### attach observer ? FUTURE
  
  ###
  @model.commit_operation
  ###
end#add_text

def dialog()
  prompts=["Text: "]
  values=["<>"]
  pops=[""]
  title="AreaTextTag"
  results=inputbox(prompts, values, pops, title)
  return nil unless results
  return results[0]
end#dialog

end#class self::TagPicked
###########################

##########################
def self.surface_data(face=nil)
  return [0, 0, [nil]] unless face && face.valid? && face.is_a?(Sketchup::Face)
  unless face.edges.any?{|e|e.soft?}
    peri=0
		face.edges.each{|e|peri+=e.length}
    return [face.area, peri, [face]]
  end
  area=0
  peri=0
  faces=self.get_surface(face)
  faces.each{|f|
    area+=f.area
		pe=0
		f.edges.each{|e|
			next if e.soft?
			next if e.faces[1] && faces.include?(e.faces[0]) && faces.include?(e.faces[1])
			pe+=e.length
		}
		peri+=pe
  }
  return [area, peri, faces]
end#surface_data
###
def self.get_surface(face)### thanks to thomthom
  surface={} ### start with hash for speed
  stack=[face]
  until stack.empty?
    stack.shift.edges.select{|e|e.soft?}.each{|edge|
      edge.faces.each{|fac|
        next if surface.key?(fac)
        stack << fac
        surface[fac]=fac
      }
    }
  end
  return surface.keys ### array
end#get_surface
######################

###########################
def self.edit()
  @model=Sketchup.active_model
  @ss=@model.selection
	ins = @ss.grep(Sketchup::ComponentInstance)[0]
  unless ins && ins.definition.get_attribute("AreaTextTag","text",nil)
    UI.messagebox("AreaTextTag:Edit\n\nYou must Preselect an AreaTextTag !")
    return nil
  end
  ### get face
  @face=ins.glued_to
  unless @face
    UI.messagebox("AreaTextTag:Edit\n\nThe Selected AreaTextTag is NOT glued onto a Face !\nRecommend deletion...")
    return nil
  end
  if ins.definition.instances[1]
    ins.make_unique
  end
  defn=ins.definition
  dents=defn.entities
  #### get settings units/dp/font/height***
  @units=defn.get_attribute("AreaTextTag","units","m2")
  @dp=defn.get_attribute("AreaTextTag","dp","2")
  @font=defn.get_attribute("AreaTextTag","font","Arial")
	### v3.0***
	@align=defn.get_attribute("AreaTextTag","align","Center")
	###
  if @model.options["UnitsOptions"]["LengthUnit"]<=1 ### it's '/"
    high=4.inch
  else
    high=100.mm
  end
  @height=defn.get_attribute("AreaTextTag","height",high)
  @surf=@model.get_attribute("AreaTextTag","surface","No")
  @model.set_attribute("AreaTextTag","surface",@surf)
  ### get areas
  area, perim, faces = self.surface_data(@face)
  ###
  area2=nil
	area2ever=false
  case @units
    when "ft2"
			area=area/144
			perim=perim/12
			punit="ft"
    when "yds2"
			area=area/1296
			perim=perim/36
			punit="yds"
    when "acres"
			area=area*0.000000159422507907
			perim=perim/12
			punit="ft"
    when "mm2"
			area=area*645.16
			perim=perim*25.4
			punit="mm"
    when "cm2"
			area=area*6.4516
			perim=perim*2.54
			punit="cm"
    when "m2"
			area=area*0.00064516
			perim=perim*0.0254
			punit="m"
    when "hectares"
			area=area*0.000000064516
			perim=perim*0.0254
			punit="m"
		when "in2"
			area=area
			perim=perim
			punit="in"
		when "m2(ft2)"
			area=area*0.00064516
			perim=perim*0.0254
			punit="m"
			area2=area*10.7639104167097223
			area2ever=true
		else
			area=0
			perim=0
			punit=""
   end
  atxt=sprintf("%."+@dp+"f", area)
  atxtu=atxt+" "+(@units.gsub(/2/,"²"))
  if area2 ### "m2(ft2)"
    atxt=sprintf("%.1f", area)
    atxtu=atxt+" m²"
		atxt2=sprintf("%.0f", area2)
		atxtu2=atxt2+" ft²"
		atxt=atxt+"\n("+atxt2+")"
		atxtu=atxtu+"\n("+atxtu2+")"
  end
	###
  text=defn.get_attribute("AreaTextTag","text","")
  ###
  prompts=["New Text: "]
  values=[text]
  pops=[""]
  title="AreaTextTag - Edit"
  results=inputbox(prompts,values,pops,title)
  return nil unless results
  text=results[0]
  ###
  if text==nil || text==""
		text=""
    txt=atxtu
  elsif text=~/[<][>]/
    txt=text.gsub(/[<][>]/, atxtu)
  elsif text=~/[<][-][>]/
    txt=text.gsub(/[<][-][>]/, atxt)
  else
    txt=text
  end
	### perimeter
	if txt=~/[|][|]/
		ptxt=sprintf("%."+@dp+"f", perim)
		ptxt<<" "+punit
		txt.gsub!(/[|][|]/, ptxt)
	elsif txt=~/[|][-][|]/
		ptxt=sprintf("%."+@dp+"f", perim)
		txt.gsub!(/[|][-][|]/, ptxt)
	end
  ###
  ### trap newlines \n
  txt.gsub!(/[\\]n/,"\n")
  #ttt=txt
  #begin
    #txt=txt.unpack("U*").map{|c|c.chr}.join
  #rescue
    #txt=ttt
  #end
	case @align
		when "Left"
			align = TextAlignLeft
		when "Right"
			align = TextAlignRight
		else ### Center
			align = TextAlignCenter
	end
  ### make changes
  @model.start_operation("AreaTextTag:Edit")
  ###
  dents.erase_entities(dents.to_a)
  dents.add_3d_text(txt, align, @font, false, false, @height, 0, 0, true, 0)
  ###
  bb=Geom::BoundingBox.new
  dents.each{|e|bb.add(e.bounds)}
  ct=bb.center
  tr=Geom::Transformation.translation(ct.vector_to([0,0,1.mm]))
  dents.transform_entities(tr, dents.to_a)
  ###
  dents.to_a.each{|e|
    if e.is_a?(Sketchup::Face)
			e.reverse! if e.normal.z<0
		elsif e.is_a?(Sketchup::Edge)
			e.hidden=true if @hidden=="Yes"
		end
		e.casts_shadows=false
    e.receives_shadows=false
  }
  defn.invalidate_bounds
  ### remember dialog's text etc as attributes
  defn.set_attribute("AreaTextTag","text",text)
  defn.set_attribute("AreaTextTag","units",@units)
  defn.set_attribute("AreaTextTag","dp",@dp)
  defn.set_attribute("AreaTextTag","font",@font)
  defn.set_attribute("AreaTextTag","align",@align)
  defn.set_attribute("AreaTextTag","height",@height)
  defn.set_attribute("AreaTextTag","hidden",@hidden)
  ### [re]attach observer ??? FUTURE
  
  ###
  @model.commit_operation
  ###
end
###########################

###########################
def self.settings()
  @model=Sketchup.active_model
  ### get settings units/dp/font/height***
  @units=@model.get_attribute("AreaTextTag","units","m2")
  @model.set_attribute("AreaTextTag","units",@units)
  @dp=@model.get_attribute("AreaTextTag","dp","2")
  @model.set_attribute("AreaTextTag","dp",@dp)
  @font=@model.get_attribute("AreaTextTag","font","Arial")
	### v3.0***
	@align=@model.get_attribute("AreaTextTag","align","Center")
  @model.set_attribute("AreaTextTag","align",@align)
	### v4.0***
	@hidden=@model.get_attribute("AreaTextTag","hidden","Yes")
  @model.set_attribute("AreaTextTag","hidden",@hidden)
	###
  @model.set_attribute("AreaTextTag","font",@font)
  if @model.options["UnitsOptions"]["LengthUnit"]<=1 ### it's '/"
    high=4.inch
  else
    high=100.mm
  end
  @height=@model.get_attribute("AreaTextTag","height",high)
  @model.set_attribute("AreaTextTag","height",@height)
  ###
  @surf=@model.get_attribute("AreaTextTag","surface","No")
  @model.set_attribute("AreaTextTag","surface",@surf)
  ###
  units=["in2",
  "ft2",
  "yds2",
  "acres",
  "mm2",
  "cm2",
  "m2",
  "hectares",
  "m2(ft2)"]
  unitsj=units.join("|")
  dps=["0",
  "1",
  "2",
  "3",
  "4",
  "5",
  "6"]
  dpsj=dps.join("|")
  fonts=["Arial",
"Bookman Old Style",
"Bradley Hand ITC",
"Century",
"Century Gothic",
"CityBlueprint",
"Comic Sans MS",
"Courier New",
"Franklin Gothic",
"Garamond",
"Gautami",
"Georgia",
"Helvetica",
"Latha",
"Papyrus",
"Rockwell",
"Romantic",
"Stylus BT",
"Tahoma",
"Times New Roman",
"Trebuchet MS",
"Verdana"
]
  fontsj=fonts.join("|")
  prompts=["Area Units: ","Text DPs : ","Text Font: ", "Text Alignment: ", "Text Height: ", "Report Surface Areas: ", "Hide Text's Edges: "]
  values=[@units,@dp,@font,@align,@height,@surf,@hidden]
  pops=[unitsj,dpsj,fontsj,"Center|Left|Right","","Yes|No", "Yes|No"]
  title="AreaTextTag - Settings"
  results=inputbox(prompts,values,pops,title)
  return nil unless results
  @units,@dp,@font,@align,@height,@surf,@hidden=results
  ###
  if @units=="m2(ft2)"
    @dp="1"
  end
  ###
  @model.set_attribute("AreaTextTag","units",@units)
  @model.set_attribute("AreaTextTag","dp",@dp)
  @model.set_attribute("AreaTextTag","font",@font)
  @model.set_attribute("AreaTextTag","align",@align)
  if @height<=0
    if @model.options["UnitsOptions"]["LengthUnit"]<=1 ### it's '/"
      @height=4.inch
    else
      @height=100.mm
    end
  end
  @model.set_attribute("AreaTextTag","height",@height)
  @model.set_attribute("AreaTextTag","surface",@surf)
  @model.set_attribute("AreaTextTag","hidden",@hidden)
  ###
  Sketchup.active_model.select_tool(self::TagPicked.new())
  ###
end
###########################

###########################
def self.update(silent=false)
  @surfaces=[]
  @model=Sketchup.active_model
  @ss=@model.selection
  if silent
    @ss.add(@model.active_entities.to_a)
  else ### NOT silent so prompt
    if @ss.empty?
      UI.beep
      ans=UI.messagebox("AreaTextTag:Update\n\nThere is NO Selection.\n\nUpdate ALL Tags in active_entities ?",MB_YESNO,"")
	  if ans==7
        return nil
	  else ### YES
	     @ss.add(@model.active_entities.to_a)
	  end
    end
  end
  ###
  tags=[]
  @ss.grep(Sketchup::ComponentInstance).each{|e|
		tags << e if e.definition.get_attribute("AreaTextTag","text",nil)
  }
	@ss.clear
	@ss.add(tags)
	###ask if update all
  if silent ### from CSV
    all=false
  else
    UI.beep
    ans=UI.messagebox("AreaTextTag:Update\n\n'YES' to update AREA ONLY.\n'NO' to update AREA and ALL SETTINGS.",MB_YESNO,"")
		if ans==6
			all=false
		else
			all=true
		end
  end
	@ss.clear
  ### make changes
  @model.start_operation("AreaTextTag:Update")
  ###
  errs=[]
  taggedfaces=[]
  dups=[]
  ###
	msg = "AreaTextTag - Updating "
	msg2=" of #{tags.length}... Please Wait..."
 tags.each_with_index{|ins, idx|
  ###
	@ss.add(ins)
	@model.active_view.refresh unless silent
	ctr = idx+1
	Sketchup::set_status_text("#{msg}#{ctr}#{msg2}", SB_PROMPT)
	###
  if ins.definition.instances[1]
    ins.make_unique
  end
  ### get face
  @face=ins.glued_to
  ###
  if taggedfaces.include?(@face)
    dups << ins
		next ### SKIP IT
  end
  taggedfaces << @face
  unless @face
   errs << ins
   next ### SKIP IT
  end
  defn=ins.definition
  dents=defn.entities
  if all
    #### get settings units/dp/font/height***
    @units=@model.get_attribute("AreaTextTag","units","m2")
    @dp=@model.get_attribute("AreaTextTag","dp","2")
    @font=@model.get_attribute("AreaTextTag","font","Arial")
		### v3.0***
		@align=@model.get_attribute("AreaTextTag","align","Center")
		@model.set_attribute("AreaTextTag","align",@align)
		###
    if @model.options["UnitsOptions"]["LengthUnit"]<=1 ### it's '/"
      high=4.inch
    else
      high=100.mm
    end
    @height=@model.get_attribute("AreaTextTag","height",high)
  else ### area only
    @units=defn.get_attribute("AreaTextTag","units","m2")
    @dp=defn.get_attribute("AreaTextTag","dp","2")
    @font=defn.get_attribute("AreaTextTag","font","Arial")
		if @model.options["UnitsOptions"]["LengthUnit"]<=1 ### it's '/"
      high=4.inch
    else
      high=100.mm
    end
    @height=defn.get_attribute("AreaTextTag","height",high)
  end
  @surf=@model.get_attribute("AreaTextTag","surface","No")
  @model.set_attribute("AreaTextTag","surface",@surf)
	@hidden=@model.get_attribute("AreaTextTag","hidden",@hidden)
  @model.set_attribute("AreaTextTag","hidden",@hidden)
  ### get areas
  area, perim, faces = self.surface_data(@face)
  ###
  area2=nil
	area2ever=false
  case @units
    when "ft2"
			area=area/144
			perim=perim/12
			punit="ft"
    when "yds2"
			area=area/1296
			perim=perim/36
			punit="yds"
    when "acres"
			area=area*0.000000159422507907
			perim=perim/12
			punit="ft"
    when "mm2"
			area=area*645.16
			perim=perim*25.4
			punit="mm"
    when "cm2"
			area=area*6.4516
			perim=perim*2.54
			punit="cm"
    when "m2"
			area=area*0.00064516
			perim=perim*0.0254
			punit="m"
    when "hectares"
			area=area*0.000000064516
			perim=perim*0.0254
			punit="m"
		when "in2"
			area=area
			perim=perim
			punit="in"
		when "m2(ft2)"
			area=area*0.00064516
			perim=perim*0.0254
			punit="m"
			area2=area*10.7639104167097223
			area2ever=true
		else
			area=0
			perim=0
			punit=""
   end
  atxt=sprintf("%."+@dp+"f", area)
  atxtu=atxt+" "+(@units.gsub(/2/,"²"))
  if area2 ### "m2(ft2)"
    atxt=sprintf("%.1f", area)
    atxtu=atxt+" m²"
		atxt2=sprintf("%.0f", area2)
		atxtu2=atxt2+" ft²"
		atxt=atxt+"\n("+atxt2+")"
		atxtu=atxtu+"\n("+atxtu2+")"
  end
  ###
  text=defn.get_attribute("AreaTextTag","text","")
  ###
  if text==nil || text==""
		text=""
    txt=atxtu
  elsif text=~/[<][>]/
    txt=text.gsub(/[<][>]/, atxtu)
  elsif text=~/[<][-][>]/
    txt=text.gsub(/[<][-][>]/, atxt)
  else
    txt=text
  end
	### perimeter
	if txt=~/[|][|]/
		ptxt=sprintf("%."+@dp+"f", perim)
		ptxt<<" "+punit
		txt.gsub!(/[|][|]/, ptxt)
	elsif txt=~/[|][-][|]/
		ptxt=sprintf("%."+@dp+"f", perim)
		txt.gsub!(/[|][-][|]/, ptxt)
	end
  ###
  ### trap newlines \n
  txt.gsub!(/[\\]n/,"\n")
  #ttt=txt
  #begin
    #txt=txt.unpack("U*").map{|c|c.chr}.join
  #rescue
    #txt=ttt
  #end
	case @align
		when "Left"
			align = TextAlignLeft
		when "Right"
			align = TextAlignRight
		else ### Center
			align = TextAlignCenter
	end
  ###
  dents.erase_entities(dents.to_a)
  dents.add_3d_text(txt, align, @font, false, false, @height, 0, 0, true, 0)
  ###
  bb=Geom::BoundingBox.new
  dents.each{|e|bb.add(e.bounds)}
  ct=bb.center
  tr=Geom::Transformation.translation(ct.vector_to([0,0,1.mm]))
  dents.transform_entities(tr, dents.to_a)
  ###
  dents.to_a.each{|e|
    if e.is_a?(Sketchup::Face)
			e.reverse! if e.normal.z<0
		elsif e.is_a?(Sketchup::Edge)
			e.hidden=true if @hidden=="Yes"
		end
		e.casts_shadows=false
    e.receives_shadows=false
  }
  defn.invalidate_bounds
  ### remember dialog's text etc as attributes
  defn.set_attribute("AreaTextTag","text",text)
  defn.set_attribute("AreaTextTag","units",@units)
  defn.set_attribute("AreaTextTag","dp",@dp)
  defn.set_attribute("AreaTextTag","font",@font)
  defn.set_attribute("AreaTextTag","align",@align)
  defn.set_attribute("AreaTextTag","height",@height)
  defn.set_attribute("AreaTextTag","hidden",@hidden)
  ### [re]attach observer ???
  
  ###
 }#tags
	@model.active_view.refresh unless silent
	Sketchup::set_status_text("#{msg} - Done.", SB_PROMPT)
  ###
	if errs[0] || dups[0]
		@ss.clear
    es=""
		es="s" if errs[1]
    ess=""
		ess="s" if dups[1]
		unless silent
			if errs[0]
				@ss.add(errs)
			 csverr="\n\n"+errs.length.to_s+" Tag"+es+" NOT glued onto a Face [shown Highlighted] !\nRecommended for Deletion..."
			else
			 csverr=""
			end
			if dups[0]
				@ss.add(dups)
				csvdup="\n\n"+dups.length.to_s+" Duplicated Face Tag"+ess+" [shown Highlighted] !\nRecommended for Deletion..."
			else
				csvdup=""
			end
			msg="AreaTextTag:Update"
			msg << csverr
			msg << csvdup
			UI.messagebox(msg)
		end
  end
	###
  @model.commit_operation
  ###
end
###########################

###########################
def self.tagSelected()
  @surfaces=[]
  @model=Sketchup.active_model
  @ents=@model.active_entities
  @ss=@model.selection
  if @ss.empty?
    UI.beep
    ans=UI.messagebox("AreaTextTag:TagSelected\n\nThere is NO Selection.\n\nTag ALL Faces in active_entities ?",MB_YESNO,"")
	if ans==7
      return nil
	else ### YES
	   @ss.add(@model.active_entities.to_a)
	end
  end
  faces=@ss.grep(Sketchup::Face)
  @ss.clear
  @ss.add(faces) if faces[0]
  unless faces[0]
    UI.messagebox("AreaTextTag:Selected\n\nThere are NO Selected Faces !")
		return nil
  end
  ### get settings units/dp/font/height***
  @units=@model.get_attribute("AreaTextTag","units","m2")
  @model.set_attribute("AreaTextTag","units",@units)
  @dp=@model.get_attribute("AreaTextTag","dp","2")
  @model.set_attribute("AreaTextTag","dp",@dp)
  @font=@model.get_attribute("AreaTextTag","font","Arial")
  @model.set_attribute("AreaTextTag","font",@font)
	### v3.0***
	@align=@model.get_attribute("AreaTextTag","align","Center")
  @model.set_attribute("AreaTextTag","align",@align)
	###
  if @model.options["UnitsOptions"]["LengthUnit"]<=1 ### it's '/"
    high=4.inch
  else
    high=100.mm
  end
  @height=@model.get_attribute("AreaTextTag","height",high)
  @model.set_attribute("AreaTextTag","height",@height)
  @surf=@model.get_attribute("AreaTextTag","surface","No")
  @model.set_attribute("AreaTextTag","surface",@surf)
  ###
  @model.start_operation("Area Text Tag")
  ###
 faces.each{|f|
  @face=f
  next if @surfaces.include?(@face)
  @surfaces << @face
  area, perim, faces = self.surface_data(@face)
	@surfaces << faces
	@surfaces.flatten!
  @surfaces.compact!
  @surfaces.uniq!
	###
	area2=nil
	area2ever=false
  case @units
    when "ft2"
			area=area/144
			perim=perim/12
			punit="ft"
    when "yds2"
			area=area/1296
			perim=perim/36
			punit="yds"
    when "acres"
			area=area*0.000000159422507907
			perim=perim/12
			punit="ft"
    when "mm2"
			area=area*645.16
			perim=perim*25.4
			punit="mm"
    when "cm2"
			area=area*6.4516
			perim=perim*2.54
			punit="cm"
    when "m2"
			area=area*0.00064516
			perim=perim*0.0254
			punit="m"
    when "hectares"
			area=area*0.000000064516
			perim=perim*0.0254
			punit="m"
		when "in2"
			area=area
			perim=perim
			punit="in"
		when "m2(ft2)"
			area=area*0.00064516
			perim=perim*0.0254
			punit="m"
			area2=area*10.7639104167097223
			area2ever=true
		else
			area=0
			perim=0
			punit=""
   end
  atxt=sprintf("%."+@dp+"f", area)
  atxtu=atxt+" "+(@units.gsub(/2/,"²"))
  if area2 ### "m2(ft2)"
    atxt=sprintf("%.1f", area)
    atxtu=atxt+" m²"
		atxt2=sprintf("%.0f", area2)
		atxtu2=atxt2+" ft²"
		atxt=atxt+"\n("+atxt2+")"
		atxtu=atxtu+"\n("+atxtu2+")"
  end
  #ttt=atxtu
  #begin
    #atxtu=atxtu.unpack("U*").map{|c|c.chr}.join
  #rescue
    #atxtu=ttt
  #end
	case @align
		when "Left"
			align = TextAlignLeft
		when "Right"
			align = TextAlignRight
		else ### Center
			align = TextAlignCenter
	end
  ###
  gp=@ents.add_group()
  gents=gp.entities
  gents.add_3d_text(atxtu, align, @font, false, false, @height, 0, 0, true, 0)
  gents.to_a.each{|e|
    if e.is_a?(Sketchup::Face)
			e.reverse! if e.normal.z<0
		elsif e.is_a?(Sketchup::Edge)
			e.hidden=true if @hidden=="Yes"
		end
		e.casts_shadows=false
    e.receives_shadows=false
  }
  ###
  ct=gp.bounds.center
  tr=Geom::Transformation.translation(ct.vector_to([0,0,1.mm]))
  gents.transform_entities(tr, gents.to_a)
  ###
  ins=gp.to_component
  ins.material="Black"
  defn=ins.definition
  dents=defn.entities
  defn.name="AreaTextTag#1"
  defn.behavior.is2d=true
  defn.behavior.snapto=SnapTo_Arbitrary
  ###
  defn.invalidate_bounds
  ###
  ins.casts_shadows=false
  ins.receives_shadows=false
  ###
  @pt=@face.bounds.center
  @pt=@pt.project_to_plane(@face.plane)
  ### relocate to pt
  tr=Geom::Transformation.new(@pt)
  ins.transform!(tr)
  ### rotate onto face
  norm=@face.normal
  if norm==Z_AXIS.reverse
    tr=Geom::Transformation.rotation(@pt, X_AXIS, 180.degrees)
    ins.transform!(tr)
  elsif norm!=Z_AXIS
    ### flatten norm
    p1=ORIGIN.offset(norm)
	p1.z=0
	fvec=ORIGIN.vector_to(p1)
    ang=fvec.angle_between(ins.transformation.yaxis.reverse)
	ang= -ang if fvec.x<0
	tr=Geom::Transformation.rotation(@pt, Z_AXIS, ang)
    ins.transform!(tr)
	### flip onto face
	ang=norm.angle_between(Z_AXIS)
	tr=Geom::Transformation.rotation(@pt, ins.transformation.xaxis, ang)
    ins.transform!(tr)
  end
  ### set gluing to face etc 
  ins.glued_to=@face
  ### remember dialog's text etc as attributes
  defn.set_attribute("AreaTextTag","text","<>")
  defn.set_attribute("AreaTextTag","units",@units)
  defn.set_attribute("AreaTextTag","dp",@dp)
  defn.set_attribute("AreaTextTag","font",@font)
  defn.set_attribute("AreaTextTag","align",@align)
  defn.set_attribute("AreaTextTag","height",@height)
  defn.set_attribute("AreaTextTag","hidden",@hidden)
  ### attach observer ? FUTURE
  
  ###
 }#faces
  ###
  @model.commit_operation
  ###
end
###########################

###########################
def self.report()
   model=Sketchup.active_model
   @surfaces=[]
   @surf=model.get_attribute("AreaTextTag","surface","No")
   model.set_attribute("AreaTextTag","surface",@surf)
   ###
   ss=model.selection
   ss.clear
   ###
   ### check model is saved...
   mpath=model.path
   if mpath.empty?
      UI.messagebox("This 'Untitled' new Model must be Saved\nbefore making an 'Area Text Tag Report' !\nExiting... ")
      return nil
   end
   path=File.dirname(mpath)
   title=model.title
   ### show VCB and status info...
   Sketchup::set_status_text(("AreaTextTag - Report... Updating..." ), SB_PROMPT)
   ###
   self.update(silent=true)
   ### show VCB and status info...
   Sketchup::set_status_text(("AreaTextTag - Report..." ), SB_PROMPT)
   ###
   entities=model.active_entities.to_a
   ###
### start undo...
   model.start_operation("AreaTextTag:Report")
### setup cdefn list...
   list=[]
   taggedfaces=[]
   nogs=[]
   dups=[]
   ###
	 area2ever=false
	 ###
   entities.grep(Sketchup::ComponentInstance).each{|e|
    next unless e.definition.get_attribute("AreaTextTag","text",nil)
	  layer=e.layer.name.tr(",",";").tr("-","_")
	  text=e.definition.get_attribute("AreaTextTag","text","").tr(",",";").gsub(/[\\]n<>/,'').gsub(/[\\]n<->/,'').gsub(/<>/,'').gsub(/<->/,'').gsub(/[\\]n/,' / ')
		text='<>' if text==''
		#ttt=text
		#begin
			#text=text.unpack("U*").map{|c|c.chr}.join
		#rescue
			#text=ttt
		#end
	  face=e.glued_to
	  if ! face
	    ss.add(e) 
		nogs << e
	    next
	  end
	  if taggedfaces.include?(face)
	    ss.add(e)
		dups << e
	    next
	  end
	  taggedfaces << face
	  next if @surfaces.include?(face)
    area=0.0
	  area2=nil
    perim=0.0
	  @surfaces << face
		area, perim, faces = self.surface_data(face)
	  @surfaces << faces
		@surfaces.flatten!
		@surfaces.compact!
	  @surfaces.uniq!
	  ###
  units=e.definition.get_attribute("AreaTextTag","units","m2")
  case units
    when "ft2"
			area=area/144
			perim=perim/12
			punit="ft"
    when "yds2"
			area=area/1296
			perim=perim/36
			punit="yds"
    when "acres"
			area=area*0.000000159422507907
			perim=perim/12
			punit="ft"
    when "mm2"
			area=area*645.16
			perim=perim*25.4
			punit="mm"
    when "cm2"
			area=area*6.4516
			perim=perim*2.54
			punit="cm"
    when "m2"
			area=area*0.00064516
			perim=perim*0.0254
			punit="m"
    when "hectares"
			area=area*0.000000064516
			perim=perim*0.0254
			punit="m"
		when "in2"
			area=area
			perim=perim
			punit="in"
		when "m2(ft2)"
			area=area*0.00064516
			perim=perim*0.0254
			punit="m"
			area2=area*10.7639104167097223
			area2ever=true
		else
			area=0
			perim=0
			punit=""
   end
   ###
	  dp=e.definition.get_attribute("AreaTextTag","dp","2")
	  area=sprintf("%."+dp+"f", area)
	  perim=sprintf("%."+dp+"f", perim)
    if area2 ### "m2(ft2)"
      area=sprintf("%.1f", area)
      area2=sprintf("%.0f", area2)
			units="m2"
    end
	  ###
	  txt=layer+","+text+","+perim+","+area+","+units
	  txt << ","+area2+",ft2" if area2
    list << txt
  }#ents
   ###
   list.sort!
   ###
   titles=["LAYER,TEXT,PERIM,AREA,UNITS"]
   if area2ever
     titles=["LAYER,TEXT,PERIM,AREA,UNITS,AREA2,UNITS2"]
   end
   list=titles+list
   ### process list..........
   ### if no tags exit...
   if ! list[0]
      UI.messagebox("There are NO 'AreaTextTags' in this Model\nto make a 'Report' !\nExiting... ")
      return nil
   end#if
### write to csv file...
    ofile=File.join(path,title+'_AreaTextTagReport.csv').tr("\\","/")
    begin
      file=File.new(ofile,"w")
    rescue### trap if open
      UI.messagebox("File:\n\n  "+ofile+"\n\nCannot be written - it's probably already open.\nClose it and try making it again...\n\nExiting...")
	  return nil
    end
    file.puts(list)
    file.close
	msg="AreaTextTag:Report:\n\n"+ofile+"\n\nWritten."
	if nogs[0]
	  es=""
	  es="s" if nogs[1]
	  msg << "\n\n"+nogs.length.to_s+" Unglued Tag"+es+" NOT Reported [shown Highlighted] !"
	end
	if dups[0]
	  es=""
	  es="s" if dups[1]
	  msg << "\n\n"+dups.length.to_s+" Duplicated Face Tag"+es+" NOT Reported [shown Highlighted] !"
	end
   UI.messagebox(msg)
   UI.openURL("file:///"+ofile)
   ###
   model.commit_operation
   ###
end#report
############

end#module
end#module
